from dragonfly_schema.window_parameter import SingleWindow, SimpleWindowRatio, \
    RepeatingWindowRatio, RepeatingWindowWidthHeight, RectangularWindows, \
    DetailedWindows
import os

# target folder where all of the samples live
root = os.path.dirname(os.path.dirname(__file__))
target_folder = os.path.join(root, 'samples')


def test_window_par_single_window():
    file_path = os.path.join(target_folder, 'window_par_single_window.json')
    SingleWindow.parse_file(file_path)


def test_window_par_simple_window_ratio():
    file_path = os.path.join(target_folder, 'window_par_simple_window_ratio.json')
    SimpleWindowRatio.parse_file(file_path)


def test_window_par_repeating_window_ratio():
    file_path = os.path.join(target_folder, 'window_par_repeating_window_ratio.json')
    RepeatingWindowRatio.parse_file(file_path)


def test_window_par_repeating_window_width_height():
    file_path = os.path.join(target_folder, 'window_par_repeating_window_width_height.json')
    RepeatingWindowWidthHeight.parse_file(file_path)


def test_window_par_detailed_rectangular_windows():
    file_path = os.path.join(target_folder, 'window_par_detailed_rectangular_windows.json')
    RectangularWindows.parse_file(file_path)


def test_window_par_detailed_windows():
    file_path = os.path.join(target_folder, 'window_par_detailed_windows.json')
    DetailedWindows.parse_file(file_path)



"""Window Parameters with instructions for generating windows."""
from pydantic import Field, root_validator, constr, conlist, confloat
from typing import List

from honeybee_schema._base import NoExtraBaseModel


class SingleWindow(NoExtraBaseModel):
    """A single window in the wall center defined by a width * height."""

    type: constr(regex='^SingleWindow$') = 'SingleWindow'

    width: float = Field(
        ...,
        gt=0,
        description='A number for the window width. Note that, if this width is '
        'applied to a wall that is too narrow for this width, the generated '
        'window will automatically be shortened when it is applied to the wall. '
        'In this way, setting the width to be `float("inf")` will create '
        'parameters that always generate a ribbon window.'
    )

    height: float = Field(
        ...,
        gt=0,
        description='A number for the window height. Note that, if this height is '
        'applied to a wall that is too short for this height, the generated '
        'window will automatically be shortened when it is applied to the wall.'
    )

    sill_height: float = Field(
        1.0,
        gt=0,
        description='A number for the window sill height.'
    )


class SimpleWindowRatio(NoExtraBaseModel):
    """A single window defined by an area ratio with the base surface."""

    type: constr(regex='^SimpleWindowRatio$') = 'SimpleWindowRatio'

    window_ratio: float = Field(
        ...,
        gt=0,
        lt=1,
        description='A number between 0 and 1 for the ratio between the window '
        'area and the parent wall surface area.'
    )


class RepeatingWindowRatio(NoExtraBaseModel):
    """Repeating windows derived from an area ratio with the base wall."""

    type: constr(regex='^RepeatingWindowRatio$') = 'RepeatingWindowRatio'

    window_ratio: float = Field(
        ...,
        gt=0,
        lt=1,
        description='A number between 0 and 1 for the ratio between the window '
        'area and the parent wall surface area.'
    )

    window_height: float = Field(
        ...,
        gt=0,
        description='A number for the target height of the windows. Note that, '
        'if the window ratio is too large for the height, the ratio will take '
        'precedence and the actual window_height will be larger than this value.'
    )

    sill_height: float = Field(
        ...,
        gt=0,
        description='A number for the target height above the bottom edge of the '
        'wall to start the windows. Note that, if the ratio is too large '
        'for the height, the ratio will take precedence and the sill_height '
        'will be smaller than this value.'
    )

    horizontal_separation: float = Field(
        ...,
        ge=0,
        description='A number for the target separation between individual window '
        'centerlines.  If this number is larger than the parent rectangle base, '
        'only one window will be produced.'
    )

    vertical_separation: float = Field(
        0,
        ge=0,
        description='An optional number to create a single vertical separation '
        'between top and bottom windows.'
    )


class RepeatingWindowWidthHeight(NoExtraBaseModel):
    """Repeating rectangular windows of a fixed width and height."""

    type: constr(regex='^RepeatingWindowWidthHeight$') = 'RepeatingWindowWidthHeight'

    window_height: float = Field(
        ...,
        gt=0,
        description='A number for the target height of the windows. Note that, '
        'if the window_height is larger than the height of the wall, '
        'the generated windows will have a height equal to the wall height '
        'in order to avoid having windows extend outside the wall face.'
    )

    window_width: float = Field(
        ...,
        gt=0,
        description='A number for the target width of the windows. '
        'Note that, if the window_width is larger than the width of the wall, '
        'the generated windows will have a width equal to the wall width '
        'in order to avoid having windows extend outside the wall face.'
    )

    sill_height: float = Field(
        ...,
        gt=0,
        description='A number for the target height above the bottom edge of the '
        'wall to start the windows. If the window_height is too large for the '
        'sill_height to fit within the rectangle, the window_height will take '
        'precedence.'
    )

    horizontal_separation: float = Field(
        ...,
        ge=0,
        description='A number for the target separation between individual window '
        'centerlines.  If this number is larger than the parent rectangle base, '
        'only one window will be produced.'
    )


class RectangularWindows(NoExtraBaseModel):
    """Several rectangular windows, defined by origin, width and height."""

    type: constr(regex='^RectangularWindows$') = 'RectangularWindows'

    origins: List[conlist(confloat(gt=0), min_items=2, max_items=2)] = Field(
        ...,
        min_items=1,
        description='An array of 2D points within the plane of the wall for the origin '
        'of each window. Each point should be a list of 2 (x, y) values. The '
        'wall plane is assumed to have an origin at the first point of the wall '
        'segment and an X-axis extending along the length of the segment. The '
        'wall plane Y-axis always points upwards. Therefore, both X and Y '
        'values of each origin point should be positive.'
    )

    widths: List[confloat(gt=0)] = Field(
        ...,
        min_items=1,
        description='An array of positive numbers for the window widths. '
        'The length of this list must match the length of the origins.'
    )

    heights: List[confloat(gt=0)] = Field(
        ...,
        min_items=1,
        description='An array of positive numbers for the window heights. '
        'The length of this list must match the length of the origins.'
    )

    @root_validator
    def check_aligned_lists(cls, values):
        "Ensure length of origins, widths and heights match."
        origins = values.get('origins')
        widths = values.get('widths')
        heights = values.get('heights')

        assert len(origins) == len(widths) == len(heights), 'Length of ' \
            'RectangularWindows origins, widths, and heights ' \
            'must match. origins: {}, widths: {}, heights: {}'.format(
                len(origins), len(widths), len(heights))

        return values


class DetailedWindows(NoExtraBaseModel):
    """Several detailed windows defined by 2D Polygons (lists of 2D vertices)."""

    type: constr(regex='^DetailedWindows$') = 'DetailedWindows'

    polygons: List[conlist(conlist(confloat(gt=0), min_items=2, max_items=2), min_items=3)] = \
        Field(
        ...,
        description='An array of arrays with each sub-array representing a polygonal '
            'boundary of a window within the plane of the wall. Each sub-array should '
            'consist of at least three 2D points and each point should be a list of '
            '2 (x, y) values. The wall plane is assumed to have an origin at the '
            'first point of the wall segment and an X-axis extending along the '
            'length of the segment. The wall plane Y-axis always points upwards. '
            'Therefore, both X and Y values of each point in the polygon should '
            'always be positive. Note that, if you are starting from 3D vertices '
            'of windows, you can use these window parameters to represent them. '
            'Some sample code to convert from 2D vertices to 2D vertices in the '
            'plane of the wall can be found here: '
            'https://www.ladybug.tools/dragonfly-core/docs/dragonfly.windowparameter.html#dragonfly.windowparameter.DetailedWindows'
    )



"""Model schema and the 3 geometry objects that define it."""
from pydantic import BaseModel, Field, root_validator, constr, conlist
from typing import List, Union

from honeybee_schema._base import IDdBaseModel
from honeybee_schema.model import Face3D, Units
from honeybee_schema.boundarycondition import Ground, Outdoors, Adiabatic, Surface

from .window_parameter import SingleWindow, SimpleWindowRatio, RepeatingWindowRatio, \
    RectangularWindows, DetailedWindows
from .shading_parameter import ExtrudedBorder, Overhang, LouversByDistance, \
    LouversByCount
from .energy.properties import Room2DEnergyPropertiesAbridged, \
    StoryEnergyPropertiesAbridged, BuildingEnergyPropertiesAbridged, \
    ContextShadeEnergyPropertiesAbridged, ModelEnergyProperties


class Room2DPropertiesAbridged(BaseModel):

    type: constr(regex='^Room2DPropertiesAbridged$') = 'Room2DPropertiesAbridged'

    energy: Room2DEnergyPropertiesAbridged = Field(
        default=None
    )


class Room2D(IDdBaseModel):

    type: constr(regex='^Room2D$') = 'Room2D'

    floor_boundary: List[conlist(float, min_items=2, max_items=2)] = Field(
        ...,
        min_items=3,
        description='A list of 2D points representing the outer boundary vertices of '
        'the Room2D. The list should include at least 3 points and each point '
        'should be a list of 2 (x, y) values.'
    )

    floor_holes: List[conlist(conlist(float, min_items=2, max_items=2), min_items=3)] = Field(
        None,
        description='Optional list of lists with one list for each hole in the floor '
        'plate. Each hole should be a list of at least 2 points and each point a list '
        'of 2 (x, y) values. If None, it will be assumed that there are no '
        'holes in the floor plate.'
    )

    floor_height: float = Field(
        ...,
        description='A number to indicate the height of the floor plane in the Z axis.'
    )

    floor_to_ceiling_height: float = Field(
        ...,
        description='A number for the distance between the floor and the ceiling.'
    )

    is_ground_contact: bool = Field(
        False,
        description='A boolean noting whether this Room2D has its floor in contact '
        'with the ground.'
    )

    is_top_exposed: bool = Field(
        False,
        description='A boolean noting whether this Room2D has its ceiling exposed '
        'to the outdoors.'
    )

    boundary_conditions: List[Union[Ground, Outdoors, Adiabatic, Surface]] = Field(
        default=None,
        description='A list of boundary conditions that match the number of segments '
        'in the input floor_geometry + floor_holes. These will be used to assign '
        'boundary conditions to each of the walls of the Room in the resulting '
        'model. Their order should align with the order of segments in the '
        'floor_boundary and then with each hole segment. If None, all boundary '
        'conditions will be Outdoors or Ground depending on whether ceiling '
        'height of the room is at or below 0 (the assumed ground plane).'
    )

    window_parameters: List[Union[None, SingleWindow, SimpleWindowRatio,
                                  RepeatingWindowRatio, RectangularWindows,
                                  DetailedWindows]] = Field(
        default=None,
        description='A list of WindowParameter objects that dictate how the window '
        'geometries will be generated for each of the walls. If None, no windows '
        'will exist over the entire Room2D.'
    )

    shading_parameters: List[Union[None, ExtrudedBorder, Overhang, LouversByDistance,
                                   LouversByCount]] = Field(
        default=None,
        description='A list of ShadingParameter objects that dictate how the shade '
        'geometries will be generated for each of the walls. If None, no shades '
        'will exist over the entire Room2D.'
    )

    properties: Room2DPropertiesAbridged = Field(
        ...,
        description='Extension properties for particular simulation engines '
        '(Radiance, EnergyPlus).'
    )

    @root_validator
    def check_segment_count(cls, values):
        "Ensure len of boundary_conditions, window par, shading par match segment count."
        floor_bound = values.get('floor_boundary')
        floor_holes = values.get('floor_holes')
        bcs = values.get('boundary_conditions')
        win_par = values.get('window_parameters')
        shd_par = values.get('shading_parameters')

        seg_count = len(floor_bound) if floor_holes is None\
            else len(floor_bound) + sum(len(hole) for hole in floor_holes)

        if bcs is not None:
            assert len(bcs) == seg_count, 'Length of Room2D boundary_conditions ' \
                'must match number of floor segments. {} != {}'.format(
                    len(bcs), seg_count)
        if win_par is not None:
            assert len(win_par) == seg_count, 'Length of Room2D window_parameters ' \
                'must match number of floor segments. {} != {}'.format(
                    len(win_par), seg_count)
        if shd_par is not None:
            assert len(shd_par) == seg_count, 'Length of Room2D shading_parameters ' \
                'must match number of floor segments. {} != {}'.format(
                    len(shd_par), seg_count)
        return values


class StoryPropertiesAbridged(BaseModel):

    type: constr(regex='^StoryPropertiesAbridged$') = 'StoryPropertiesAbridged'

    energy: StoryEnergyPropertiesAbridged = Field(
        default=None
    )


class Story(IDdBaseModel):

    type: constr(regex='^Story$') = 'Story'

    room_2ds: List[Room2D] = Field(
        ...,
        description='An array of dragonfly Room2D objects that together form an '
        'entire story of a building.'
    )

    floor_to_floor_height: float = Field(
        None,
        description='A number for the distance from the floor plate of this story '
        'to the floor of the story above this one (if it exists). If None, this '
        'value will be the maximum floor_to_ceiling_height of the input room_2ds.'
    )

    multiplier: int = Field(
        1,
        ge=1,
        description='An integer that denotes the number of times that this Story is '
        'repeated over the height of the building.'
    )

    properties: StoryPropertiesAbridged = Field(
        ...,
        description='Extension properties for particular simulation engines '
        '(Radiance, EnergyPlus).'
    )


class BuildingPropertiesAbridged(BaseModel):

    type: constr(regex='^BuildingPropertiesAbridged$') = 'BuildingPropertiesAbridged'

    energy: BuildingEnergyPropertiesAbridged = Field(
        default=None
    )


class Building(IDdBaseModel):

    type: constr(regex='^Building$') = 'Building'

    unique_stories: List[Story] = Field(
        ...,
        description='An array of unique dragonfly Story objects that together form '
        'the entire building. Stories should generally be ordered from lowest '
        'floor to highest floor. Note that, if a given Story is repeated several '
        'times over the height of the building, the unique story included in this '
        'list should be the first (lowest) story of the repeated floors.'
    )

    properties: BuildingPropertiesAbridged = Field(
        ...,
        description='Extension properties for particular simulation engines '
        '(Radiance, EnergyPlus).'
    )


class ContextShadePropertiesAbridged(BaseModel):

    type: constr(regex='^ContextShadePropertiesAbridged$') = \
        'ContextShadePropertiesAbridged'

    energy: ContextShadeEnergyPropertiesAbridged = Field(
        default=None
    )


class ContextShade(IDdBaseModel):

    type: constr(regex='^ContextShade$') = 'ContextShade'

    geometry: List[Face3D] = Field(
        ...,
        description='An array of planar Face3Ds that together represent the '
        'context shade.'
    )

    properties: ContextShadePropertiesAbridged = Field(
        ...,
        description='Extension properties for particular simulation engines '
        '(Radiance, EnergyPlus).'
    )


class ModelProperties(BaseModel):

    type: constr(regex='^ModelProperties$') = 'ModelProperties'

    energy: ModelEnergyProperties = Field(
        default=None
    )


class Model(IDdBaseModel):

    type: constr(regex='^Model$') = 'Model'

    buildings: List[Building] = Field(
        ...,
        description='A list of Buildings in the model.'
    )

    context_shades: List[ContextShade] = Field(
        None,
        description='A list of ContextShades in the model.'
    )

    units: Units = Field(
        default=Units.meters,
        description='Text indicating the units in which the model geometry exists. '
        'This is used to scale the geometry to the correct units for simulation '
        'engines like EnergyPlus, which requires all geometry be in meters.'
    )

    tolerance: float = Field(
        default=0.01,
        ge=0,
        description='The maximum difference between x, y, and z values at which '
        'vertices are considered equivalent. This value should be in the Model '
        'units and is used in a variety of checks and operations. A value of 0 '
        'will result in bypassing all checks so it is recommended that this always '
        'be a positive number when checks have not already been performed '
        'on a Model. The default of 0.01 is suitable for models in meters.'
    )

    angle_tolerance: float = Field(
        default=1.0,
        ge=0,
        description='The max angle difference in degrees that vertices are '
        'allowed to differ from one another in order to consider them colinear. '
        'This value is used in a variety of checks and operations that can be '
        'performed on geometry. A value of 0 will result in no checks and '
        'an inability to perform certain operations so it is recommended that '
        'this always be a positive number when checks have not already '
        'been performed on a given Model.'
    )

    properties: ModelProperties = Field(
        ...,
        description='Extension properties for particular simulation engines '
        '(Radiance, EnergyPlus).'
    )



"""Run all sample-genrating modules."""
import os
import importlib


# gather all of the modules the contain sample files
root_dir = os.path.dirname(os.path.abspath(__file__))
module_names = os.listdir(root_dir)
module_files = [os.path.join(root_dir, mod) for mod in module_names]
sample_modules = []
for mod, mod_file in zip(module_names, module_files):
    if os.path.isfile(mod_file) and mod_file.endswith('.py') \
            and mod.startswith('sample'):
        sample_modules.append(mod[:-3])

# execute each of the sample file modules
for mod in sample_modules:
    print('Generating samples for {}'.format(mod))
    importlib.import_module(mod)
